from time import sleep

from adafruit_pixelbuf import PixelBuf

from pixelbuf_pi_animation.data import Animation


class SimplePixelBufPlayer:
    """
    This class is able to take Animation objects and render them via any class that inherits from Adafruit's Pixelbuf
    class.

    Adafruit publishes a Python library called
    [Adafruit_CircuitPython_Pixelbuf](https://github.com/adafruit/Adafruit_CircuitPython_Pixelbuf). This library serves
    as the base class for many of their LED drivers. `PixelBufPlayer` is able to take anything that inherits from
    `PixelBuf` and use it to render output.

    An example of this is using the DotStar driver to handle the actual rendering:

    ```
    import adafruit_dotstar
    import board


    num_leds = 11
    controller = adafruit_dotstar.DotStar(board.SCLK, board.MOSI, num_leds, pixel_order=adafruit_dotstar.BGR, auto_write=False)
    animation = ...
    player = PixelBufPlayer(controller)
    player.load(animation)
    player.play()
    ```

    This player is the simple and most naive implementation you could imagine. Since the Pi is not a real-time computer,
    there are no guarantees around timing made. It also makes no attempt to correct for these things as the animation
    plays on. That will come in another release.
    """

    def __init__(self, controller: PixelBuf):
        """
        Constructor for creating SimplePixelBufPlayer objects. It takes an Adafruit driver that is built on the
        `adafruit_pixelbuf` library.
        :param controller: an Adafruit driver
        """
        self._controller: PixelBuf = controller
        self._animation: Animation = None

    def load(self, animation: Animation) -> None:
        """
        Pass the player the animation to play.


        :param animation: the animation to play
        :return: None
        """
        # TODO Are there any sanity checks we wish to run here?
        self._animation = animation

    def play(self) -> None:
        """
        Renders the frames to the controller device.

        Before calling this method, the user must call load() and pass in the Animation. Once that is done, the user can
        call this method to start the rendering. This function will parse and read the next frame, send it to the
        LED controller, and wait the specified amount of time before moving on.

        :return: None
        """
        # TODO We probably want to implement some sort of interrupt system here

        if not self._animation:
            raise RuntimeError('The load() function must be called before playing an animation can happen.')

        loop_count = 0
        while True:
            # TODO We need some sort of timing to help us with figuring out slow-downs so we know when to skip frames
            # Overall animation loop
            for frame_idx, frame in enumerate(self._animation.frames):

                # Loops the frames
                for pixel_idx, pixel in enumerate(frame.pixels):

                    # Loops the pixels
                    # TODO There is an efficiency that can possibly be gained here by "parsing" these frames earlier
                    if pixel is None:
                        self._controller[pixel_idx] = (0, 0, 0, 1.0)
                    else:
                        self._controller[pixel_idx] = (pixel.red, pixel.blue, pixel.green, pixel.brightness)

                # Show the frame for a while
                self._controller.show()
                sleep(frame.display_ms / 1000)

            # The animation is complete; do some book-keeping after
            sleep(self._animation.pause_between_play_ms / 1000)
            if self._animation.loop_infinitely:
                continue

            loop_count += 1
            if loop_count >= self._animation.max_plays:
                break

        self._controller.fill((0, 0, 0))
        self._controller.show()



import functools
from dataclasses import dataclass
from typing import List, Optional


@dataclass
class Pixel:
    """
    This class represents a single pixel within some Frame. One or more pixels together make up a Frame.
    """

    brightness: float
    """Represents the brightness of a pixel; must be between 0.0 and 1.0, inclusive."""

    red: int
    """Represents the red value of a pixel; must be between 0 and 255, inclusive."""

    green: int
    """Represents the green value of a pixel; must be between 0 and 255, inclusive."""

    blue: int
    """Represents the blue value of a pixel; must be between 0 and 255, inclusive."""

    def __post_init__(self):
        self._validate_brightness()
        self._validate_color()

    def _validate_brightness(self):
        min_b = 0.0
        max_b = 1.0
        if self.brightness < min_b or self.brightness > max_b:
            raise ValueError(f'Brightness must be between {min_b} and {max_b}, inclusive. '
                             f'The brightness for this pixel was set to {self.brightness}')

    def _validate_color(self):
        colors = {
            'red': self.red,
            'green': self.green,
            'blue': self.blue,
        }
        min_c = 0
        max_c = 255

        for color, value in colors.items():
            if value < min_c or value > max_c:
                raise ValueError(f'Colors must be between {min_c} and {max_c}, inclusive. '
                                 f'The color {color} for this pixel was set to {value}')


@dataclass
class Frame:
    """
    This class represents a single frame of animation. One or more Frames stitched together create an Animation.
    """
    display_ms: int
    """The amount of time this frame should be displayed for, in milliseconds"""

    pixels: List[Pixel]
    """A list of Pixels that compose this frame, in the order that they are connected to the strand"""

    def __post_init__(self):
        self._validate_display_ms()
        self._validate_pixels()

    def _validate_display_ms(self):
        min_d = 1
        if self.display_ms < min_d:
            raise ValueError(f'Display ms must be a positive integer and not less than {min_d}. '
                             f'The display for this frame was set to {self.display_ms}')

    def _validate_pixels(self):
        min_p = 1
        if len(self.pixels) < min_p:
            raise ValueError(f'There must be a least {min_p} pixel(s) within this frame.')


@dataclass
class Animation:
    """
    This class represents a collection of frames that are to be played as an animation
    """

    frames: List[Frame]
    """A list of Frames in the order that they should be rendered"""

    loop_infinitely: bool
    """Determines if this animation should be shown in an infinite loop"""

    pause_between_play_ms: int
    """How many milliseconds to pause between animation loops; will be applied after showing all frames"""

    max_plays: Optional[int]
    """How many times to show the animation; Must be set to `None` if `loop_infinitely` is True"""

    def __post_init__(self):
        self._validate_frames()
        self._validate_pause_between_play_ms()
        self._validate_max_plays()

    def __len__(self):
        return len(self.frames)

    @property
    def frame_total_time_ms(self) -> int:
        """
        This property loops through all of the frames, sums the delay on each one, and returns the total.
        :return: total expected delay, in ms, of this animation
        """
        return functools.reduce(lambda value, element: value+element.display_ms, self.frames, 0)

    def _validate_frames(self):
        min_f = 1
        if len(self.frames) < min_f:
            raise ValueError(f'There must be at least {min_f} frame(s) within this animation.')

    def _validate_pause_between_play_ms(self):
        if self.pause_between_play_ms < 0:
            raise ValueError(f'The pause between plays cannot be negative.')

    def _validate_max_plays(self):
        if self.max_plays and self.max_plays < 0:
            raise ValueError(f'The max plays cannot be negative.')

        if self.max_plays and self.loop_infinitely:
            raise ValueError('max_plays must be set to None if loop_infinitely is set to True')




import abc
import dataclasses
import json
from typing import IO

import dacite
import msgpack

from pixelbuf_pi_animation.data import Animation


class AnimationParser(abc.ABC):
    """
    Base class for all animation parsers.

    This base class defines all methods that are required for implementing an animation parser.
    """

    @staticmethod
    @abc.abstractmethod
    def read(file_input: IO) -> Animation:
        """
        Given some sort of file input, this method will read that file and return an Animation object.
        :param file_input:
        :return: an Animation object
        """
        pass

    @staticmethod
    @abc.abstractmethod
    def write(writable_file: IO, animation: Animation) -> None:
        """
        Given a writable file and an animation, this method will write the given animation into that file. This method
        does not handle file cleanup or closing.
        :param writable_file:
        :param animation:
        :return:
        """
        pass


class JsonAnimationParser(AnimationParser):
    """
    Defines how to read Animations to/from JSON.
    """

    @staticmethod
    def read(file_input: IO) -> Animation:
        parsed_dict = json.load(file_input)
        return dacite.from_dict(data_class=Animation, data=parsed_dict)

    @staticmethod
    def write(writable_file: IO, animation: Animation) -> None:
        json.dump(dataclasses.asdict(animation), writable_file)


class MsgPackAnimationParser(AnimationParser):
    """
    Defines how to read/write animations in a compact MessagePack binary format.
    """

    @staticmethod
    def read(file_input: IO) -> Animation:
        parsed_dict = msgpack.unpack(file_input)
        return dacite.from_dict(data_class=Animation, data=parsed_dict)

    @staticmethod
    def write(writable_file: IO, animation: Animation) -> None:
        msgpack.pack(dataclasses.asdict(animation), writable_file)
